
##
## This file provides a Python/numpy implementation of the benchmark
## collection from the research paper
## "Challenges of Convex Quadratic Bi-objective Benchmark Problems"
## Proceedings of the Genetic and Evolutionary Computation Conference
## (ACM), 2019, by Tobias Glasmachers.
##
## Copyright 2024 Tobias Glasmachers
##
## Permission is hereby granted, free of charge, to any person obtaining a
## copy of this software and associated documentation files (the
## "Software"), to deal in the Software without restriction, including
## without limitation the rights to use, copy, modify, merge, publish,
## distribute, sublicense, and/or sell copies of the Software, and to
## permit persons to whom the Software is furnished to do so, subject to
## the following conditions:
##
## The above copyright notice and this permission notice shall be included
## in all copies or substantial portions of the Software.
##
## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
## OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
## IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
## CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
## TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
## SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
##

import numpy as np

class MOBenchmark:
	"""
	Quadratic bi-objective problem.
	"""
	def __init__(self, name, dimension, instance, kappa = 1e3):
		"""
		Construct a function form name, search space dimension, a numeric
		problem instance id, and a parameter controlling ill-conditioning.
		"""
		self.name = name
		self.__name__ = name
		self.dimension = dimension
		self.instance = instance
		self.kappa = kappa

		self.a1 = 1
		self.b1 = 0
		self.x1 = np.zeros(dimension)
		self.U1 = np.zeros((dimension, dimension))
		self.D1 = np.zeros(dimension)
		self.A1 = np.zeros((dimension, dimension))
		self.H1 = np.zeros((dimension, dimension))
		self.a2 = 1
		self.b2 = 0
		self.x2 = np.zeros(dimension)
		self.U2 = np.zeros((dimension, dimension))
		self.D2 = np.zeros(dimension)
		self.A2 = np.zeros((dimension, dimension))
		self.H2 = np.zeros((dimension, dimension))
		self.delta = np.zeros(dimension)
		self.s = 1
		rng = np.random.default_rng(instance)

		# parse the name
		if len(name) != 3: raise ValueError("invalid problem name: " + name)
		category = int(name[0])
		if category < 1 or category > 9: raise ValueError("invalid problem name: " + name)
		if name[1] != '|' and name[1] != '/': raise ValueError("invalid problem name: " + name)
		self.aligned = name[1] == '|'
		if name[2] == 'C': self.s = 1
		elif name[2] == 'I': self.s = 0.5
		elif name[2] == 'J': self.s = 0.25
		else: raise ValueError("invalid problem name: " + name)

		# create the problem instance
		deltaFromGEV = False

		if category == 1:
			self.U1 = np.eye(dimension)
			self.U2 = np.eye(dimension)
			self.D1 = np.ones(dimension)
			self.D2 = np.ones(dimension)
			if self.aligned:
				self.delta[rng.integers(dimension)] = 1
			else:
				self.delta = rng.normal(size=dimension)
				self.delta /= np.linalg.norm(self.delta)
		elif category == 2:
			self.U1 = np.eye(self.dimension)
			self.U2 = np.eye(self.dimension)
			self.D1 = np.ones(dimension)
			if self.aligned:
				self.D2 = self._createD(rng)
				self.delta[rng.integers(dimension)] = 1
			else:
				i = rng.integers(dimension)
				j = rng.integers(dimension - 1)
				if j >= i: j += 1
				self.D2 = self._createDdup(rng, i, j)
				angle = 2 * np.pi * rng.random()
				self.delta[i] = np.cos(angle)
				self.delta[j] = np.sin(angle)
		elif category == 3:
			self.U1 = np.eye(self.dimension)
			self.U2 = np.eye(self.dimension)
			if self.aligned:
				self.D1 = self._createD(rng)
				self.delta[rng.integers(dimension)] = 1
			else:
				i = rng.integers(dimension)
				j = rng.integers(dimension - 1)
				if j >= i: j += 1
				self.D1 = self._createDdup(rng, i, j)
				angle = 2 * np.pi * rng.random()
				self.delta[i] = np.cos(angle)
				self.delta[j] = np.sin(angle)
			self.D2 = self.D1
		elif category == 4:
			self.U1 = np.eye(self.dimension)
			self.U2 = np.eye(self.dimension)
			if self.aligned:
				self.D1 = self._createD(rng)
				self.D2 = self._createD(rng)
				self.delta[rng.integers(dimension)] = 1
			else:
				i = rng.integers(dimension)
				j = rng.integers(dimension - 1)
				if j >= i: j += 1
				self.D1 = self._createDdup(rng, i, j)
				self.D2 = self._createDdup(rng, i, j)
				angle = 2 * np.pi * rng.random()
				self.delta[i] = np.cos(angle)
				self.delta[j] = np.sin(angle)
		elif category == 5:
			self.U1 = np.eye(self.dimension)
			self.D1 = np.ones(dimension)
			self.D2 = self._createD(rng)
			if self.aligned:
				i = rng.integers(dimension)
				self.U2 = self._sampleUfix(rng, i)
				self.delta[i] = 1.0
			else:
				self.U2 = self._sampleU(rng)
				deltaFromGEV = True
		elif category == 6:
			self.U1 = np.eye(self.dimension)
			self.D1 = self._createD(rng)
			self.D2 = self._createD(rng)
			if self.aligned:
				i = rng.integers(dimension)
				self.U2 = self._sampleUfix(rng, i)
				self.delta[i] = 1.0
			else:
				self.U2 = self._sampleU(rng)
				deltaFromGEV = True
		elif category == 7:
			self.D1 = self._createD(rng)
			self.D2 = self.D1
			if self.aligned:
				i = rng.integers(dimension)
				self.U1 = self._sampleUfix(rng, i)
				self.delta[i] = 1.0
			else:
				self.U1 = self._sampleU(rng)
				deltaFromGEV = True
			self.U2 = self.U1
		elif category == 8:
			self.D1 = self._createD(rng)
			self.D2 = self._createD(rng)
			if self.aligned:
				i = rng.integers(dimension)
				self.U1 = self._sampleUfix(rng, i)
				self.delta[i] = 1.0
			else:
				self.U1 = self._sampleU(rng)
				deltaFromGEV = True
			self.U2 = self.U1
		elif category == 9:
			self.U1 = self._sampleU(rng)
			self.U2 = self._sampleU(rng)
			self.D1 = self._createD(rng)
			self.D2 = self._createD(rng)
			if self.aligned:
				V, W = self._eig(self.U1, self.D1, self.U2, self.D2)
				i = rng.integers(dimension)
				delta = V[:, i]
				delta /= np.linalg.norm(delta)
				UT = self._sampleUTdelta(rng, delta)
				self.U1 = UT @ self.U1
				self.U2 = UT @ self.U2
				self.delta = UT @ delta
			else:
				deltaFromGEV = True
		else: assert(False)   # internal error, should have been caught earlier

		self.A1 = self.U1 @ np.diag(np.sqrt(self.D1))
		self.A2 = self.U2 @ np.diag(np.sqrt(self.D2))
		self.H1 = self.A1 @ np.transpose(self.A1)
		self.H2 = self.A2 @ np.transpose(self.A2)

		if deltaFromGEV:
			V, W = self._eig(self.U1, self.D1, self.U2, self.D2)
			i = rng.integers(dimension)
			self.delta = V[:, i]
			self.delta /= np.linalg.norm(self.delta)

		## sample single-objective optima in the range [-5, 5]^n
		center = rng.normal(size = dimension)
		while np.max(np.abs(center)) >= 4.5:
			center = rng.normal(size = dimension)
		self.x1 = center - 0.5 * self.delta
		self.x2 = center + 0.5 * self.delta

		self.a1 = pow(10.0, 6 * rng.random())
		self.a2 = pow(10.0, 6 * rng.random())
		self.b1 = 2 * self.a1 * rng.random() - self.a1
		self.b2 = 2 * self.a2 * rng.random() - self.a2

	def __call__(self, x):
		"""evaluate the function"""
		d1 = np.transpose(self.A1) @ (x - self.x1)
		d2 = np.transpose(self.A2) @ (x - self.x2)
		return np.array([
				0.5 * self.a1 * np.power(d1@d1, self.s) + self.b1,
				0.5 * self.a2 * np.power(d2@d2, self.s) + self.b2,
			])

	def utopian(self):
		"""This information is for evaluation purposes, it shall not be made available to the optimizer"""
		return np.array([self.b1, self.b2])

	def nadir(self):
		"""This information is for evaluation purposes, it shall not be made available to the optimizer"""
		d1 = np.transpose(self.A1) @ self.delta;
		d2 = np.transpose(self.A2) @ self.delta;
		return np.array([
				0.5 * self.a1 * np.power(d1@d1, self.s) + self.b1,
				0.5 * self.a2 * np.power(d2@d2, self.s) + self.b2,
			])

	def _createD(self, rng):
		ret = np.power(self.kappa, np.arange(self.dimension) / (self.dimension - 1))
		rng.shuffle(ret)
		return ret

	def _createDdup(self, rng, u, v):
		if v < u: u,v = v,u
		z = self.dimension - 1
		ret = np.zeros(self.dimension)
		ret[:-1] = np.power(self.kappa, np.arange(z) / (z-1.0))
		ret[z] = ret[u]
		if v != z: ret[v],ret[z] = ret[z],ret[v]
		return ret

	def _sampleU(self, rng):
		n = self.dimension
		ret = rng.normal(size = (n, n))
		for i in range(n):
			for j in range(i):
				ret[i] -= (ret[i] @ ret[j]) * ret[j]
			ret[i] /= np.linalg.norm(ret[i])
		return ret

	def _sampleUfix(self, rng, u):
		n = self.dimension
		ret = rng.normal(size = (n, n))
		ret[u, :] = 0
		ret[:, u] = 0
		ret[u, u] = 1
		for i in range(n):
			for j in range(i):
				ret[i] -= (ret[i] @ ret[j]) * ret[j]
			ret[i] /= np.linalg.norm(ret[i])
		return ret

	def _sampleUTdelta(self, rng, delta):
		n = self.dimension
		ret = rng.normal(size = (n, n))
		ret[0] = delta
		# Gram-Schmidt
		for i in range(n):
			for j in range(i):
				ret[i] -= (ret[i] @ ret[j]) * ret[j]
			ret[i] /= np.linalg.norm(ret[i])
		i = rng.integers(n)
		if i != 0:
			ret[[0, i], :] = ret[[i, 0], :]
		return ret

	def _eig(self, U1, D1, U2, D2):
		invsqrtD2 = np.power(D2, -0.5)
		H1 = U1 @ np.diag(D1) @ np.transpose(U1)
		invA2 = U2 @ np.diag(invsqrtD2) @ np.transpose(U2)
		M = invA2 @ H1 @ invA2
		val, vec = np.linalg.eigh(M)
		V = invA2 @ vec
		return V, val
